#ifdef WIN32
#define _CRT_SECURE_NO_DEPRECATE
#endif
#include "tclap/CmdLine.h"	//	TCLAP Include  (must be top one...otherwise compile errors)
#include <sys/stat.h>
#include <string>			// C++ strings

extern "C" {
	#include "cluster.h" /* The C Clustering Library */
}

#include <cfloat> 
#include <cassert>
#include <cmath>
#include "VocabularyTree.h"
#include "TemplateDatabase.h"

typedef unsigned char       BYTE;
using namespace TCLAP;
using namespace std;

bool ReadKeyFileDesc(char *filename, unsigned char** data, unsigned int**);
bool ReadDesc(FILE *fp, unsigned char** data, unsigned int**);
bool FileExists(string strFilename);
bool FileExists(const char * strFilename);
void fillTree(VocabularyTree& tree, unsigned char** pointList, double** ctrs, int** cmask, int nPts, unsigned char* clusterid, double* weight);
void kmeansRecursion(VocabularyTree& tree, unsigned char** pointList, double** ctrs, int** cmask, int nPts, unsigned char* clusterid, double* weight, unsigned char level, long node);
unsigned int* findPathIndex(VocabularyTree& tree, unsigned char* queryDesc);
unsigned int findOnLevel(VocabularyTree& tree, unsigned char* queryDesc, int level, unsigned int childIndex);
double getEDistance(unsigned char* queryVector, unsigned char* nodeVector);
int addImageToDatabase(VocabularyTree& tree, TemplateDatabase& database, unsigned int imgID);
int updateDatabase(char *filename, VocabularyTree& tree, TemplateDatabase& database, unsigned int imgID);


//----------------------------------------------------------------------
//  Global parameters (some are set in getArgs())
//----------------------------------------------------------------------

int numk = 10;				// number of centers
int	dim	= 128;				// dimension
int numLevels = 6;			// number of levels
int numImages = 5000;
int	stages = 100;			// number of stages
int nrows = 5000;			//	nrows >= number of keypoints expected
int npass = 8;
int sample_count = 0;		//	Final count of keypoints read

bool FileExists (string strFilename)
{
	return FileExists(strFilename.c_str());
}

bool FileExists(const char * strFilename)
{
	struct stat stFileInfo;
	bool blnReturn;
	int intStat;

	// Attempt to get the file attributes
	intStat = stat(strFilename,&stFileInfo);
	if(intStat == 0) {
		// We were able to get the file attributes
		// so the file obviously exists.
		blnReturn = true;
	} 
	else 
	{
		// We were not able to get the file attributes.
		// This may mean that we don't have permission to
		// access the folder which contains this file. If you
		// need to do that level of checking, lookup the
		// return values of stat which will give you
		// more details on why stat failed.
		blnReturn = false;
	}

	return(blnReturn);
}

/*
	Return true if there was an error reading a file.
	Return false if there was no error.
*/
bool ReadKeyFileDesc(char *filename, unsigned char **data, unsigned int **pos)
{
	FILE *file;
	file = fopen (filename, "r");
	if (! file)
	{
		printf("Could not open file: %s", filename);
        fclose( file );
		return true;
	}

	return ReadDesc(file, data, pos);
}

/*
	Return true if there was an error reading a file.
	Return false if there was no error.
*/
bool ReadDesc(FILE *fp, unsigned char **data, unsigned int **pos)
{
	int num, len;
	//unsigned char val;
	int val;
	float row, col, scale, ori;

	if (fscanf(fp, "%d %d", &num, &len) != 2)
	{
		printf("Invalid keypoint file beginning.\n");
        fclose( fp );
		return true;
	}

	if (len != 128)
	{
		printf("Keypoint descriptor length invalid (should be 128).\n");
		fclose( fp );
        return true;
	}

	for(int descCount = 0; descCount  < num; descCount++)
	{
		if(sample_count == nrows)
			break;

		if (fscanf(fp, "%f %f %f %f", &row, &col, &scale,
			&ori) != 4)
		{
			printf("Invalid keypoint file format.\n");
			fclose( fp );
            return true;
		}

		pos[descCount][0] = (unsigned int) col;
		pos[descCount][1] = (unsigned int) row;

		for (int i = 0; i < len; i++)
		{
			if (fscanf(fp, "%d", &val) != 1 || val < 0 || val > 255)
			{
				printf("Invalid keypoint file value.\n");
                fclose( fp );
				return true;
			}

			data[sample_count][i] = (unsigned char)val;
		}
		
		sample_count++;
		/*tempCount = descCount;*/
	}

	/*sample_count += descCount;*/
	fclose( fp );
    return false;
}

unsigned int* findPathIndex(VocabularyTree& tree, unsigned char* queryDesc)
{
	unsigned int *index = new unsigned int[numLevels];
	unsigned int childIndex = 0;

	for(int ii = 0; ii < numLevels; ii++)
		index[ii] = 0;

	for(int tempLevel = 0; tempLevel < numLevels; tempLevel++)
	{
		index[tempLevel] = findOnLevel(tree, queryDesc, tempLevel+1, childIndex+1);
		childIndex = tree.getChildArrayIndex(tempLevel+1, index[tempLevel] + 1, 1);
	}

	return index;
}

//return is zero indexed
unsigned int findOnLevel(VocabularyTree& tree, unsigned char* queryDesc, int level, unsigned int childIndex){
	double eDistance = DBL_MAX, newEDistance = 0;
	unsigned int index = 0;

	for(int ii = 0; ii < numk; ii++){
		if(tree.nodeExist(level, childIndex + ii))
		{
			newEDistance = getEDistance(queryDesc, tree.getDescVector(level, childIndex + ii));

			if(newEDistance < eDistance){
				eDistance = newEDistance;
				index = ii;
			}
		}
		else
			//return (ii - 1 - index) + childIndex - 1;
			break;
	}

	return index + childIndex - 1;
}

double getEDistance(unsigned char* queryVector, unsigned char* nodeVector)
{
	double distance = 0;

	for(int element = 0; element < dim; element++)
	{
		distance += pow(((double)(nodeVector[element]) - (double)(queryVector[element])),2);
	}

	return sqrt(distance);
}

void fillTree(VocabularyTree& tree, unsigned char** pointList, double** ctrs, int** cmask, int nPts, unsigned char* clusterid, double* weight)
{
	kmeansRecursion(tree, pointList, ctrs, cmask, nPts, clusterid, weight, 1, 0);
}

void kmeansRecursion(VocabularyTree& tree, unsigned char** pointList, double** ctrs, int** cmask, int nPts, unsigned char* clusterid, double* weight, unsigned char level, long node)
{
	if(level <= numLevels)
	{
		bool nextLevelExists = false, numPointsCheck = false;

		if(	(level + 1)	<= numLevels)
			nextLevelExists = true;

		if(nPts >= numk)
			numPointsCheck = true;

		//put the current level centers into the tree.
		for(long nodeIndex = 0 + node; (nodeIndex - node) < numk; nodeIndex++)
		{
			if((nodeIndex - node) < nPts){
				//cout << "**********************Level " << (int)level << " Node " << nodeIndex << endl;
				tree.allocateNode(level, nodeIndex + 1);
				for(int column = 0; column < dim; column++)
				{
					assert(cmask[nodeIndex - node][column] != 0); // if this fails, a center element is missing
					tree.setDescVectorElement(level, nodeIndex + 1, column + 1, (unsigned char)(floor(ctrs[nodeIndex - node][column]+.5)));
					//cout << (int)tree.getDescVectorElement(level, nodeIndex + 1, column +1) << " ";
				}

				//cout << endl << endl;

			}
			else
				break;
		}

		unsigned char*** points;

		if(nextLevelExists && numPointsCheck)
		{
			points = new unsigned char**[numk];
			int *descClusCount = new int[numk];
			int *refRow = new int[numk];
			memset(descClusCount, 0, sizeof(int)*numk);	//	initialize to zero
			memset(refRow, 0, sizeof(int)*numk);		//	initialize to zero

			for(int row = 0; row < nPts; row++)
				descClusCount[clusterid[row]]++;		//count for descriptor vectors in each cluster.

			for(int i = 0; i < numk; i++)
				points[i] = new unsigned char*[descClusCount[i]];//allocate descriptor vector list.

			for(int row = 0; row < nPts; row++)
			{
				points[clusterid[row]][refRow[clusterid[row]]] = pointList[row];
				refRow[clusterid[row]]++;
			}

			delete [] clusterid;

			for(int i = 0; i < numk; i++)
			{
				if(level == 1) cout << "Level 1, Cluster " << i+1 << " in progress...";
				
				/*cout << "level " << int(level) << " cluster " << i+1 << " Number of points " << descClusCount[i] << endl;
				for(int temp = 0; temp < descClusCount[i]; temp++)
					for(int temp2 = 0; temp2 < dim; temp2++)
						cout << points[i][temp][temp2] << " ";

				cout << endl;*/

				int tempk = numk;

				if(tempk > descClusCount[i]) {// numk cannot be greater than the number of points
					tempk = descClusCount[i];
				}

					unsigned char* tempClusterid = new unsigned char[descClusCount[i]];
					memset(tempClusterid, 0, sizeof(unsigned char)*descClusCount[i]);
					double error = 0;
					int ifound = 0;

					double** cdata = new double*[tempk];
					int** tempCmask = new int*[tempk];

					for (int temp = 0; temp < tempk; temp++){
						cdata[temp] = new double[dim];
						memset(cdata[temp], 0, sizeof(double)*dim);	//	initialize to zero
						tempCmask[temp] = new int[dim];
						memset(tempCmask[temp], 0, sizeof(int)*dim);	//	initialize to zero
					}

					kcluster(tempk, descClusCount[i], dim, points[i], weight,
						0, npass, 'a', 'e', tempClusterid, &error, &ifound);

					getclustercentroids(tempk, descClusCount[i], dim, points[i], tempClusterid,
						cdata, tempCmask, 0, 'a');

					long clusIndex = tree.getChildArrayIndex(level, (node+1) + i, 1);

					kmeansRecursion(tree, points[i], cdata, tempCmask, descClusCount[i], tempClusterid, weight, level+1 , clusIndex);

					for (int temp = 0; temp < tempk; temp++)
					{ free(cdata[temp]);
					free(tempCmask[temp]);
					}
					free(cdata);
					free(tempCmask);

				if(level == 1)
					cout << "DONE" << endl;
			}

			delete[] descClusCount;
			delete[] refRow;
		}
		else if(nextLevelExists){

			long clusIndex = tree.getChildArrayIndex(level, (node+1), 1);
			unsigned char* tempClusterid = NULL;
			unsigned char** tempPointList = NULL;
			double* tempWeight = NULL;

			for(int temp = 0; temp < nPts; temp++)
				kmeansRecursion(tree, tempPointList, &ctrs[temp], &cmask[temp], 1, tempClusterid, tempWeight, level+1, clusIndex);
		}
	}
}

int addImageToDatabase(VocabularyTree& tree, TemplateDatabase& database, unsigned int imgID)
{
	//IplImage *img = cvCreateImage(cvSize(ImageMaxX,ImageMaxY), IPL_DEPTH_8U, 1);

	//char imgName[256];
	//sprintf(imgName, "N:\\Shared\\Scalable Image Recognition Code\\TemplateImages\\%d.jpg",imgID);
	//sprintf(imgName, "N:\\Shared\\Scalable Image Recognition Code\\Small Image Database\\ukbench (%d).jpg",imgID);

	char tempName[256];
	sprintf(tempName, "%d.key",imgID);

	//char systemCall[256];
	//sprintf(systemCall, "siftWin32.exe <tmp.pgm >%d.key",imgID);

	//if(FileExists(imgName))
	if(FileExists(tempName))
	{
		//img = cvLoadImage( imgName, 0 );

		//pgmb_write( "tmp.pgm" , ImageMaxX, ImageMaxY, (unsigned char *)img->imageData );

		//cvReleaseImage(&img);

		//system(systemCall);

		if(updateDatabase(tempName, tree, database, imgID) == -1)
		{
			return -1;
		}
	}
	else
	{
		cout << "******************************************************" << endl;
		cout << "Image not added to database!!" << endl;
		cout << "No features found!!" << endl;
		cout << "Check if the images exists at the directory specified." << endl;
		cout << "******************************************************" << endl;
		//cvReleaseImage(&img);
		return 0;
	}

	//cvReleaseImage(&img);
	return 1;
}

int updateDatabase(char *filename, VocabularyTree& tree, TemplateDatabase& database, unsigned int imgID)
{
	FILE *file;
	file = fopen (filename, "r");

	char saveKeyName[256];
	sprintf(saveKeyName, "%d.key",imgID);

	if (! file)
	{
		printf("Could not open key file: %s", filename);
		cout << endl << "Database not updated!!!!" << endl;
		fclose( file );
		return -1;
	}

	int num, len, val;
	float row, col, scale, ori;
	unsigned char tempDesc[128];

	if (fscanf(file, "%d %d", &num, &len) != 2)
	{
		printf("Invalid keypoint file beginning.\n");
		cout << "Database not updated!!!!" << endl;
		fclose( file );
		return -1;
	}

	if (len != 128)
	{
		printf("Keypoint descriptor length invalid (should be 128).\n");
		cout << "Database not updated!!!!" << endl;
		fclose( file );
		return -1;
	}

	if (num == 0){
		cout << "Image " << imgID << " not added to database. No descriptor vectors found." << endl;
		return -1;
	}

	database.incramentNumberOfImages();

	for(int descCount = 0; descCount < num; descCount++)
	{
		if (fscanf(file, "%f %f %f %f", &row, &col, &scale,
			&ori) != 4)
		{
			printf("Invalid keypoint file format.\n");
			cout << "Database not updated!!!!" << endl;
			fclose( file );
			return -1;
		}

		for (int i = 0; i < len; i++)
		{
			if (fscanf(file, "%d", &val) != 1 || val < 0 || val > 255)
			{
				printf("Invalid keypoint file value.\n");
				cout << "Database not updated!!!!" << endl;
				fclose( file );
				return -1;
			}

			tempDesc[i] = (unsigned char)val;
		}

		unsigned int *path = findPathIndex(tree, tempDesc);

		database.allocateInvertedFile(path[numLevels-1]);
		database.addImageToInvertedFile(imgID, path[numLevels-1]);
		database.updateForwardFileTree(imgID, path[numLevels-1]);


	}

	fclose( file );

	/*char ch;
	ifstream infile;
	ofstream outfile;
	filebuf *inbuf, *outbuf;

	infile.open (filename, ios::in);
	outfile.open (saveKeyName, ios::out);

	inbuf = infile.rdbuf();
	outbuf = outfile.rdbuf();

	ch = inbuf->sgetc();
	while ( ch != EOF)
	{
		outbuf->sputc(ch);
		ch = inbuf->snextc();
	}

	outfile.close();
	infile.close();*/

	return 0;
}
