#include "defs.h"

int main( int argc, char** argv )
{
	//	Begin parsing the command line arguments
	try {
		VocabularyTree tree;
		TemplateDatabase database;

		CmdLine cmd("This application is a demo to illustrate how to use the classes provided to create a vocabulary tree using keyfiles, add objects to a database, and query objects in the database.", ' ', "1.0.0", true);

		ValueArg<int> numberlevels("l", string("numLevels"), string("Specify the number of levels."), false, 6, string("6"), cmd);
		ValueArg<int> numclusters("k", string("numberClusters"), string("Specify the number of centers."), false, 10, string("10"), cmd);
		ValueArg<int> dimensions("d", string("dimensions"), string("Vector dimension."), false, 128, string("128"), cmd);
		ValueArg<int> numberimages("i", string("numImages"), string("Number of images (or keys) to build the tree from."), false, 5000, string("5000"), cmd);
		ValueArg<string> binTree("b", string("binTree"), string("Path to the location of a binary tree file."), false, string("C:\\L6C10STbin.txt"), string("C:\\L6C10STbin.txt"), cmd);
		ValueArg<string> db("a", string("database"), string("Path to the location of the database."), false, string("database.txt"), string("database.txt"), cmd);
		ValueArg<string> keysPath("p", string("keysPath"), string("Path to the location of the key files.  With formatting in printf style for loading file names."), false, string("C:\\VocabularyTreeKeys\\frame%d.key"), string("C:\\VocabularyTreeKeys\\frame%d.key"), cmd);
		ValueArg<string> newKey("n", string("newImage"), string("Path to the location of a key file that represents a new image."), false, string("C:\\newImage.txt"), string("C:\\newImage.txt"), cmd);
		ValueArg<string> queryKey("q", string("queryImage"), string("Path to the location of a key file that represents an image to query from the database."), false, string("C:\\newImage.txt"), string("C:\\newImage.txt"), cmd);

		cmd.parse(argc, argv);	//	Parse the command line arguments

		numLevels = numberlevels.getValue();
		numk = numclusters.getValue();
		dim = dimensions.getValue();
		numImages = numberimages.getValue();

		//	Approximate the database size, set the database size, allocate memory for it
		unsigned long int numInvFile = (unsigned long int)pow((double)numk, (double)numLevels);
		database.setDatabase(numInvFile, 0);
		database.allocatePointerArray();

		if(db.isSet())
			database.loadDatabase(const_cast<char*>(db.getValue().c_str()));

		//	Load the binary tree, if it's location was given
		if(binTree.isSet()) {
			tree.binaryModeRetrieveTreeIntoMem(const_cast<char*>(binTree.getValue().c_str()));
			printf("%s has been read into memory.\n", binTree.getValue().c_str());
		} else {	//	Otherwise initialize and create an empty tree
			tree.setDimension(dim);
			tree.setLevel(numLevels);
			tree.setNumberOfClusters(numk);
			if(tree.allocateTree() == 1)
				cout << "Tree allocated!" << endl;
			else
				cout << "Tree not initialized" << endl;
		}

		//	If there was a path set to keys and there is no binary tree
		//	then try and load key files and build a vocabulary tree
		if(keysPath.isSet() && !binTree.isSet()) {
			bool done = false;
			int keyNum = 0;
			char keyName[256];
			nrows = numImages * 1000;		//	Estimate array size

			unsigned char **tempdata = new unsigned char*[nrows];	//	Where the keypoints are loaded to
			unsigned int **pos = new unsigned int*[nrows];		//	Don't know why?
			for(int ii = 0; ii < nrows; ii++)	{
				tempdata[ii] = new unsigned char[dim];
				pos[ii] = new unsigned int[2];
			}

			printf("Loading %d key files.\n", numImages);

			//	While not done continue the loop
			while(!done && keyNum <= numImages) {
				sprintf(keyName, keysPath.getValue().c_str(), keyNum);	//	Create file name
				done = !FileExists(keyName);	//	Check to see if file exists
				if(!done) done = ReadKeyFileDesc(keyName, tempdata, pos);	//	If file exists load data and check for errors
				if(!done) printf("Success...%s\r", keyName);
				else printf("Error...%s\n", keyName);
				keyNum++;
			}
			
			printf("Shrinking array from %d to %d, down by %d bytes.\r", nrows, sample_count, sizeof(unsigned char)*(nrows - sample_count));
			unsigned char **data = new unsigned char*[sample_count];	//	Where the keypoints are loaded to
			for(int ii = 0; ii < sample_count; ii++) {
				data[ii] = new unsigned char[dim];
				memcpy(data[ii], tempdata[ii], sizeof(unsigned char) * dim);
			}

			for(int ii = 0; ii < nrows; ii++) {
				delete [] tempdata[ii];
				delete [] pos[ii];
			}

			delete [] tempdata;
			tempdata = NULL;
			delete [] pos;
			pos = NULL;

			nrows = sample_count;		//	Adjust the array size to be only the size of the number of keypoints actually loaded

			printf("Loaded %d keys, which had %d keypoints.\t\t\t\t\n", keyNum - 1, sample_count);


			printf("Please wait while we allocate memory.\r");
			unsigned char *clusterid = new unsigned char[nrows];//	The cluster number to which each item belongs.
			memset(clusterid, 0, sizeof(unsigned char) * nrows);//	Initialize the array to zero
			double **centers = new double*[nrows];				//	This matrix stores the centroid information.
			int **centermask = new int*[nrows];					//	This matrix stores which values in centers are missing.
			double *weight = new double[nrows];					//	The weights that are used to calculate the distance.
			memset(weight, 0, sizeof(double)*nrows);			//	Initialize the array to zero
			double error = 0;
			int ifound = 0;
			for(int ii = 0; ii < nrows; ii++)	{
				centers[ii] = new double[dim];
				memset(centers[ii], 0, sizeof(double)*dim);		//	Fill with zeros
				centermask[ii] = new int[dim];
				memset(centermask[ii], 0, sizeof(int)*dim);		//	Fill with zeros
			}
			printf("Memory has been allocated for the calculations.\n");

			printf("Clustering the data.\n");
			kcluster(numk, nrows, dim, data, weight, 0, npass, 'a', 'e', clusterid, &error, &ifound);

			//	See the The C Clustering Library for more detail...generate centers of data
			printf("Generating the centers of the data.\n");
			if(getclustercentroids(numk, nrows, dim, data, clusterid, centers, centermask, 0, 'a') == 0)
				cerr << "** There was an error generating the cluster centroids. **" << endl;

			//	Create the tree
			printf("Please wait while we fill the tree.\n");
			fillTree(tree, data, centers, centermask, sample_count, clusterid, weight);

			//	Output the tree file as a binary file
			char outputName[256];
			sprintf(outputName, "BinaryTree_%dL_%dC.bin", numLevels, numk);
			tree.binaryModeWriteTreeFile(outputName);
			//tree.textModeWriteTreeFile(outputName);	//	If you want human readable output...
			
			//	Delete the array of data from loading the keys
			for(int ii = 0; ii < nrows; ii++) {
				delete [] data[ii];
				delete [] centers[ii];
				delete [] centermask[ii];
			}

			delete [] data;
			delete [] clusterid;
			delete [] centers;
			delete [] centermask;
			delete [] weight;
		}

		//	If there was a new key passed at the command line,
		//	then we will try and add that key to the database
		if(newKey.isSet()) {
			if(FileExists(newKey.getValue())) {
				updateDatabase(const_cast<char*>(newKey.getValue().c_str()), tree, database, database.N + 1);	// Add to database (N + 1, is length plus 1)
				database.saveDatabase();	//	Save changes to the database
			} else
				printf("The key file cannot be found.  Thus it will not be added to the database.\n\t%s\n", newKey.getValue().c_str());

		}

		//	If there was a query request made then we will try
		//	and query the database for the image that was passed in.
		if(queryKey.isSet()) {
			if(FileExists(queryKey.getValue())) {
				unsigned char **tempPts = new unsigned char*[nrows];
				unsigned int **tempPos = new unsigned int*[nrows];
				for(int i = 0; i < nrows; i++) {
					tempPts[i] = new unsigned char[dim];
					tempPos[i] = new unsigned int[2];
				}


				sample_count = 0;
				ReadKeyFileDesc(const_cast<char*>(queryKey.getValue().c_str()), tempPts, tempPos);
				for(int i = 0; i < sample_count; i++) {
					unsigned int *path = findPathIndex(tree, tempPts[i]);
					database.updateQueryTree(path[numLevels-1], i);
					delete path;
					path = NULL;
				}

				database.getMatch();

				//	Print scoring results
				for(int ii = 0; ii < database.N; ++ii)
					printf("Image id %d has a matching score of: %d\n", ii, database.getScore(ii));

			} else 
				printf("The key file cannot be found.  Thus it will not be searched for in the database.\t%s\n", queryKey.getValue().c_str());
		}

	} catch (ArgException &e) {
		cerr << "error: " << e.error() << " for arg " << e.argId() << endl;
	}

	system("PAUSE");	//	Wait...

	return 0;
}